'''
This code is part of Cosmic Commando; Copyright (C) 2012 Piotr 'ZasVid' Sikora; see CosmicCommando.py for full notice.
'''

from game_pieces import *
from game_tileboard import *
import libtcodpy as libtcod
from utils import *
from guts.game_piece_defs import *

class Rect(object):
    #a rectangle on the map. used to characterize a room.
    def __init__(self, x, y, width, height):
        self.x = x
        self.y = y
        self.x_end = x + width
        self.y_end = y + height
        self.w = width
        self.h = height
    
    def center(self):
        center_x = (self.x + self.x_end) / 2
        center_y = (self.y + self.y_end) / 2
        return (center_x, center_y)
 
    def intersect(self, other):
        #returns true if this rectangle intersects with another one
        return (self.x <= other.x_end and self.x_end >= other.x and
                self.y <= other.y_end and self.y_end >= other.y)


def generate_board(width, height, max_rooms, room_min_size, room_max_size, player):
    generator = BoardGenerator(width, height, max_rooms, room_min_size, room_max_size)
    generator.generate()
    generator.place_player(player)
    return generator.board

class BoardGenerator(object):
    
    def __init__(self, board_width, board_height, max_rooms, room_min_size, room_max_size):
        self.board = Board(board_width, board_height)
        self.max_rooms = max_rooms
        self.room_min_size = room_min_size
        self.room_max_size = room_max_size
    
    def generate(self):
        self.build_board()
        self.populate_board()
    
    def create_room(self, room):
        #go through the tiles in the rectangle and make them passable
        for x in range(room.x + 1, room.x_end):
            for y in range(room.y + 1, room.y_end):
                self.board[x,y].terrain = new_terrain('floor')

    def create_h_tunnel(self, x1, x2, y):
        for x in range(min(x1, x2), max(x1, x2) + 1):
            self.board[x,y].terrain = new_terrain('floor')

    def create_v_tunnel(self, y1, y2, x):
        #vertical tunnel
        for y in range(min(y1, y2), max(y1, y2) + 1):
            self.board[x,y].terrain = new_terrain('floor')

    def build_board(self):

        for x in range(self.board.width):
            for y in range(self.board.height):
                self.board[x,y].terrain = new_terrain('wall')
        rooms = []
        num_rooms = 0
        first_done = False
        for _ in range(self.max_rooms):
            if first_done:
                #random width and height
                w = random_int(self.room_min_size, self.room_max_size)
                h = random_int(self.room_min_size, self.room_max_size)
            else:
                w = 10
                h = 10
            #random position without going out of the boundaries of the map
            x = random_int(0, self.board.width - w - 1)
            y = random_int(0, self.board.height - h - 1)
            #"Rect" class makes rectangles easier to work with
            new_room = Rect(x, y, w, h)
            #run through the other rooms and see if they intersect with this one
            failed = False
            for other_room in rooms:
                if new_room.intersect(other_room):
                    failed = True
                    break
            if not failed:
                #this means there are no intersections, so this room is valid
                #"paint" it to the map's tiles
                self.create_room(new_room)
                #center coordinates of new room, will be useful later
                (new_x, new_y) = new_room.center()
                if num_rooms == 0:
                    self.player_position = (new_x,new_y)
                else:
                    #all rooms after the first:
                    #connect it to the previous room with a tunnel
                    #center coordinates of previous room
                    (prev_x, prev_y) = rooms[num_rooms-1].center()
                    #draw a coin (random number that is either 0 or 1)
                    if coinflip():
                        #first move horizontally, then vertically
                        self.create_h_tunnel(prev_x, new_x, prev_y)
                        self.create_v_tunnel(prev_y, new_y, new_x)
                    else:   
                        #first move vertically, then horizontally
                        self.create_v_tunnel(prev_y, new_y, prev_x)
                        self.create_h_tunnel(prev_x, new_x, new_y)
                #finally, append the new room to the list
                rooms.append(new_room)
                #add some contents to this room, such as monsters
                num_rooms += 1    
        self.rooms = rooms

    def populate_board(self):
        #first room done by hand for testing
        first_room = self.rooms[0]
        frcx, frcy = first_room.center()
        
        px, py = first_room.x + 2, first_room.y + 2
        for direction in eight_directions:
            self.board[px + direction[0], py + direction[1]].put(new_item('grenade'))
        
        #features first
        for room in self.rooms:
            for x in range(room.x, room.x_end):
                if not self.board[x, room.y].is_blocked():
                    self.board[x, room.y].feature = new_feature('closed door')
                if not self.board[x, room.y_end].is_blocked():
                    self.board[x, room.y_end].feature = new_feature('closed door')
            for y in range(room.y, room.y_end):
                if not self.board[room.x, y].is_blocked():
                    self.board[room.x, y].feature = new_feature('closed door')
                if not self.board[room.x_end, y].is_blocked():
                    self.board[room.x_end, y].feature = new_feature('closed door')
        #enemies! we need enemies!
        for room in self.rooms:
            x,y = room.center()
            if not self.board[x,y].is_blocked():
                if coinflip() or room == self.rooms[0]:
                    self.board[x,y].occupant = new_agent('Raven officer')
                    acceptable_directions = []
                    for direction in compass_directions:
                        dx, dy = directions[direction]
                        if not self.board[x + dx, y + dy].is_blocked():
                            acceptable_directions.append(direction)
                    bodyguard_spawns = random_combination(acceptable_directions, min(len(acceptable_directions), random_nat(1)+2))
                    for spawn_point in bodyguard_spawns:
                        dx, dy = directions[spawn_point]
                        self.board[x + dx, y + dy].occupant = new_agent('Raven recruit')
                else:
                    self.board[x,y].occupant = new_agent('cybork')
        for i in range(1,10):
            x,y = directions[i]
            if not self.board[frcx + x, frcy + y].is_blocked():
                self.board[frcx + x, frcy + y].feature = new_feature('closed door', "test door")
                break 

        #items to use         
        for room in self.rooms[1:]:
            self.board[room.x + 1, room.y + 1].put(new_item('medkit'))

    def place_player(self, player):
        first_room = self.rooms[0]
        (player_x, player_y) = first_room.x + 2, first_room.y + 2
        self.board[player_x, player_y].occupant = player
