'''
This code is part of Cosmic Commando; Copyright (C) 2012 Piotr 'ZasVid' Sikora; see CosmicCommando.py for full notice.
'''

import shelve, dbhash
import libtcodpy as libtcod

class GameData():
    """ Game data stores all the data that is generated for the actual game.
        Part of the data is primary - that is, it is generated procedurally 
        at the beginning of the game or by player's choice or actions and
        therefore needs to be stored in a file when saving a game.
        Then, there's secondary data, that can be regenerated from the primary
        data, by simple operations (done here) or using external methods 
        - the latter data must be passed to the object of this class 
        as parameters to ensure that this module is the root dependency
        of the project.
        
    """  
    
    def __init__(self):
        """ Saving and loading data needs to make sure it doesn't store
            any object twice independently (e.g. player as a piece on 
            the board and then self.player handle), because shelve will
            NOT keep these connected. However, some pieces of the data
            might be annoying to recreate procedurally, so we can index
            them for ease of use.  
        """
        self.new()
        
    def new(self):
        #Primary data:
        self.msgs = []
        self.msg_archive = []
        self.time_limited_turns = False
        self.turncount = 0
        self._board = None
        self.state = None
        #Secondary data - for storing and loading via index
        self.player = None
        #Secondary data - for procedural recreation
        self.fov_recompute = None
        self.fov_map = None
        
    def _get_pieces(self):
        pieces = []
        for tile in self.board.tiles:
            pieces.extend(tile.pieces)
        return pieces
    pieces = property(_get_pieces)
    
    def _get_agents(self):
        agents = []
        for tile in self.board.tiles:
            if tile.occupant is not None:
                agents.append(tile.occupant)
        return agents
    agents = property(_get_agents)
    
    def _get_board(self):
        return self._board
    def _set_board(self, board):
        self.fov_recompute = None
        self.fov_map = None
        self._pieces = None
        self._agents = None
        self._board = board
    board = property(_get_board, _set_board)
    
    def set_fov_at(self, x, y):
        libtcod.map_set_properties(self.fov_map, x, y, not self.board[x,y].blocks_sight, not self.board[x,y].is_blocked)
      
    def in_fov(self, x, y):
        #TODO: isn't it better to use game.board[x,y].visible?
        return libtcod.map_is_in_fov(self.fov_map, x, y)
        
    def save(self, savegame):
        """ If the same object is shelved two times (e.g. by itself and as a part of
            a collection, weird things happen. Make sure it doesn't happen.
        """
        #open a new empty shelve (possibly overwriting an old one) to write the game data
        savefile = shelve.open(savegame, 'n')
        #primary data
        savefile['board'] = self.board
        savefile['messages'] = self.msgs
        savefile['message archive'] = self.msg_archive
        savefile['game state'] = self.state
        savefile['time limited turns'] = self.time_limited_turns
        #secondary data 
        savefile['player position'] = (self.player.x, self.player.y)
        savefile.close()
        
    def load(self, savegame):
        self.new()
        #open the previously saved shelve and load the game data
        savefile = shelve.open(savegame, 'r')
        #primary data
        self.board = savefile['board'] 
        self.msgs = savefile['messages'] 
        self.msg_archive = savefile['message archive'] 
        self.state = savefile['game state'] 
        self.time_limited_turns = savefile['time limited turns'] 
        #secondary data
        player_x, player_y = savefile['player position']
        self.player = game.board[player_x, player_y].occupant 
        
        savefile.close()
        self.prepare_secondary_data()
    
    def prepare_secondary_data(self):
        #TODO: regenerate
        pass 

game = GameData()