'''
This code is part of Cosmic Commando; Copyright (C) 2012 Piotr 'ZasVid' Sikora; see CosmicCommando.py for full notice.
'''

import os
import libtcodpy as libtcod
from guts.messages import message
from guts.debug import debug
from guts.game_data import game
from guts.game_piece_defs import new_player
from guts.board_generation import generate_board
from guts.player_input import InputInterpreter 
from guts.config import MSG_WIDTH
from guts.utils import compass_directions, directions, distance


class GamePlay(object):
    '''
    This class handles playing the game.
    '''

    def __init__(self, ui, achievements):
        self.ui = ui
        self.achievements = achievements
        self.input_interpreter = InputInterpreter()
        self.savefile = 'savegame'

    def new(self):
        #create object representing the player
        game.player = new_player()
        #generate map (at this point it's not drawn to the screen)
        game.board = generate_board(width = 75, height = 75, max_rooms = 15, room_min_size = 3, room_max_size = 10, player = game.player)
        game.state = 'playing'
        #create the list of game messages and their colors, starts empty
        game.msgs = []
        message('Welcome soldier! The fate of the galaxy rests in your hands.', 'red')

    def load(self, savefile):
        game.load(savefile)
        
    def play(self):
        self.prepare_play()
        turns_taken = 0
        center_on_player = True
        self.refresh = True
        while not libtcod.console_is_window_closed():
            #render the screen
            if self.refresh:
                self.ui.render_all(center_on_player)
                self.refresh = False
            else:
                self.ui.render_mouseover_info()
            libtcod.console_flush()
            if turns_taken == 0 or game.state != 'playing':
                turns_taken, quit_game, center_on_player = self.player_action(center_on_player)
            if game.state == 'playing' and turns_taken > 0:
                turns_taken -= 1
                for agent in game.agents:
                    if agent != game.player and agent['ai'] is not None:
                        agent.act()
            if quit_game:
                if game.state == 'dead':
                    self.achievements.add_highscore(0)
                    self.achievements.remember(game)
                    #os.remove(self.savefile) TODO: enable to disable savescumming. Possibly make a debug option
                    game.state = 'finished'
                else:
                    game.save(self.savefile)
                break #leave this play
        if game.state != 'finished':
            game.save(self.savefile)

    def select_target_tile(self, occupied = False, max_range = None, in_fov = True):
        #return a tile left-clicked in player's FOV (optionally in a range),
        #or (None,None) if right-clicked or escaped.
        message('Left click on a{0} to select a target. <Esc>, <Space>, or Right Click to cancel.'.format('n enemy' if occupied else ' tile'))
        while True:
            #TODO: refresh stuff under mouse and hud
            self.ui.render_hud()
            libtcod.console_flush()
            player_action = self.input_interpreter.command(self.ui.get_input())
            #print(str(player_action))
            if player_action == 'TARGET':
                tile = self.ui.get_tile_under_mouse()
                fov_constraint = not in_fov or game.in_fov(tile.x, tile.y)
                range_constraint = max_range is None or distance(game.player.x, game.player.y, tile.x, tile.y) <= max_range
                occupied_constraint = not occupied or tile.occupant is not None #TODO: and not player?  
                if fov_constraint and range_constraint and occupied_constraint:
                    if occupied:
                        return tile.occupant
                    else:
                        return tile
                elif not fov_constraint:
                    message("Target must be visible!")
                elif not range_constraint:
                    message("Target outside range of {0} tiles.".format(max_range))
                elif not occupied_constraint:
                    message("Target must be a combatant!")
            elif player_action in ['CANCEL', 'QUIT']:
                return None
            
    def player_action(self, old_center_on_player):
        """ Returns how many turns passed, whether the player has quit
            and whether the camera should center on player.
            #TODO: consider a dictionary of player action outcomes.
        """
        center_on_player = old_center_on_player
        player_action = self.input_interpreter.command(self.ui.get_input())
        if player_action is None:
            return (0, False, center_on_player)
        player = game.player        
        center_on_player = True
        self.refresh = True
        """
            TODO: REALTIME decisions for hasty players
            if not game.turn_mode: 
                if self.action_timeout == 0:
                    self.action_timeout = GAME_ACTION_TIMEOUT
                    return 'waiting'
                self.action_timeout -= 1
        """
        if game.state == 'playing':
            """ To avoid excessive returning group the actions:
                1st group uses up a turn
                2nd group after else doesn't.
            """
            turns_taken = 1 #default action length is one, but lets not get constrained by it!
            if player_action == 'QUIT':
                return (0, True, center_on_player)
            if player_action.startswith('DIRECTION '):
                self.player_act_in_direction(directions[player_action[10:]])
            elif player_action == 'INTERACT':
                """ This allows to interact with features that can't be bumped into.
                """
                traversible_features = []
                x,y = player.x, player.y
                for tile in game.board.ring(x,y):
                    if tile.feature is not None and not tile.is_blocked():
                        traversible_features.append((tile.feature,
                                                    directions['full names'][(tile.x -x, tile.y - y)]))
                if len(traversible_features) > 1:
                    self.ui.msgbox("Too many to choose!") #TODO: let them choose
                elif len(traversible_features) == 1:
                    traversible_features[0][0].interact(player)
                else:
                    message("Nothing to interact with.")
                    turns_taken = 0
            elif player_action.startswith('PICK UP'):
                #TODO: menu to choose a pickup if more than one
                tile = game.board[player.x, player.y] 
                if tile.items:
                    for item in [item for item in tile.items]:
                        player.pick_up(item)
                else:
                    message("Nothing to pick up.")
            elif player_action.startswith('DROP'):
                chosen_item = self.inventory_menu('Press the key next to an item to drop it, or any other to cancel.\n', ['droppable'])
            elif player_action.startswith('INVENTORY'):
                chosen_item = self.inventory_menu('Press the key next to an item to equip it, or any other to cancel.\n')
                if chosen_item is not None:
                    player.equip(chosen_item)
            elif player_action.startswith('USE'):
                action_type = player_action[4:]
                item = player[action_type + ' item']
                if item is not None:
                    if item[action_type].requires_target:
                        message("You take aim with your {0}...".format(item.name))
                        target = self.select_target_tile(occupied = (action_type == 'ranged'))
                        if target is None:
                            message("Use of " + item.name + " cancelled.")
                            return (0, False, center_on_player)
                        item.use(game.player, action_type, target)
                    else:
                        item.use(game.player, action_type)
                else:
                    turns_taken = 0
                    message("Nothing to use in " + player_action[4:] + " slot!")
            elif player_action.startswith('TARGET'):
                #TODO: move tile 
                tile = self.ui.get_tile_under_mouse()
                if tile.occupant is not None and player['ranged item'] is not None and tile.visible:
                    player['ranged item'].use(game.player, 'ranged', tile.occupant)
                else:
                    message("Shoot only at enemies, ricochets are dangerous!")
                    return (0, False, center_on_player)
            else :
                if player_action.startswith('CAMERA '):
                    dx, dy = directions[player_action[7:]]
                    self.ui.move_camera(dx * 10, dy * 10)
                    center_on_player = False
                elif player_action == 'DISPLAY MESSAGES':
                    msgs = [msg[0] for msg in game.msg_archive]
                    self.ui.msgbox('\n'.join(msgs), width = MSG_WIDTH)
                elif player_action == 'DISPLAY CHARACTER INFO':
                    self.ui.msgbox('TODO: Display character info.')
                elif player_action == 'HELP':
                    self.ui.msgbox('Keybindings: \n' \
                                   'Movement: \n' \
                                   '     [q] [w] [e] [a] [d] [z] [x] [c]\n' \
                                   '     numpad digits\n' \
                                   '     arrow keys (only orthogonally)\n' \
                                   'Wait: numpad 5 or [s]\n' \
                                   'Move Camera: ctrl + arrow keys \n' \
                                   'Save & Quit: ESC\n' \
                                   'Close doors: Space\n' \
                                   'Pick up items: [g]\n' \
                                   'Open inventory: [i]\n' \
                                   'Shoot: [r] or LMB on an enemy\n' \
                                   'Use equipped medkit: [v]\n' \
                                   'Throw equipped grenade: [f]\n' \
                                   'Activate equipped lightning launcher: [tab]\n' \
                                   'Show this help: [?]'
                                   )
                return (0, False, center_on_player)
            return (turns_taken, False, center_on_player)
        elif game.state == 'dead':
            if player_action in ['QUIT', 'INTERACT']:
                return (0, True, center_on_player)
        return (0, False, center_on_player) #TODO: does it make sense

    def inventory_menu(self, header, filters = []):
        #show a menu with each item of the inventory as an option
        filter_set = set(filters)
        filtered_inventory = [item for item in game.player.inventory if filter_set.issubset(set(item.flags))]
        if len(filtered_inventory) == 0:
            options = ['Nothing applicable.']
            #TODO: make it a msgboxy inventory
        else:
            options = []
            for item in filtered_inventory:
                option = item.name 
                for flag in ['ranged', 'melee', 'consumable', 'deployable', 'activated']:
                    if game.player[flag + ' item'] == item:
                        option += ' (equipped)' 
                options.append(option)
        index = self.ui.menu(header, options)
        #if an item was chosen, return it
        if index is None or len(filtered_inventory) == 0: return None
        return filtered_inventory[index]
    
    def player_act_in_direction(self, direction):
        """ Move or attack or perform other contextual actions.
            TODO: move to Agent class?
        """
        dx, dy = direction
        #the coordinates the player is moving to/attacking
        x = game.player.x + dx
        y = game.player.y + dy
        if not game.board.contains_position(x,y):
            message.add("It is the edge of the world!")
        #try to find an attackable object there
        target_tile = game.board[x,y]
        if target_tile.is_blocked():
            if target_tile.occupant is not None:
                if target_tile.occupant.is_player():
                    message("You wait.")
                else:
                    game.player.melee(target_tile.occupant)
            elif target_tile.feature is not None:
                target_tile.feature.interact(game.player)
            else:
                message("You can't go there!")
        else:
            game.player.move_to(x, y)
            game.fov_recompute = True
    
    def prepare_play(self):
        """ This method:
            1) computes the necessary game data that cannot be handled by 
            game_data module to avoid circular dependencies 
            2) sets other things necessary for the game to be played that
            it would be a waste to do elsewhere.  
        """
        self.initialize_fov()
        for piece in game.pieces:
            if piece.has('health') and not piece.has('max health'):
                piece['max health'] = piece['health']
            if piece.has('shields') and not piece.has('max shields'):
                piece['max shields'] = piece['shields']
            
    def initialize_fov(self):
        game.fov_recompute = True
        board = game.board
        game.fov_map = libtcod.map_new(board.width, board.height)
        for y in range(board.height):
            for x in range(board.width):
                game.set_fov_at(x, y)
        if not debug.fog:  #debug mode, reveal everything
            for y in range(board.height):
                for x in range(board.width):
                    board[x,y].explored = True
            for piece in game.pieces:
                piece.always_visible = True

        